# TPK4186 - Advanced Tools for Performance Engineering
# TextInterface.py
# by Antoine Rauzy
# Copyrights (c) 2019-2020 NTNU
#
# This file is part of the educational material provided to NTNU students registered to the course TPK4186.
# This work is licensed under a Creative Commons Attribution-NonCommercial 4.0 International License.
# For details about this license, see https://creativecommons.org/licenses/by-nc/4.0/.

# Table of Contents
# -----------------
# 1. Required modules
# 2. Tokens
# 3. Exceptions
# 4. Lexer
# 5. Parser
# 6. Checker
# 7. Printer

# 1. Required modules
# -------------------

import sys
import re
from MarkovModels import *

# 2. Tokens
# ---------

class Token:
  LPAR = 1
  RPAR = 2
  COLON = 3
  SEMICOLON = 4
  COMMA = 5
  EQUAL = 6
  PLUS = 7
  MINUS = 8
  STAR = 9
  SLASH = 10
  NUMBER = 11
  IDENTIFIER = 12
  KEYWORD_MODEL = 13
  KEYWORD_END = 14
  KEYWORD_STATE = 15
  KEYWORD_INITIAL = 16
  KEYWORD_TRANSITION = 17
  KEYWORD_PARAMETER = 18
  KEYWORD_REWARD = 19

  def __init__(self, type, string, lineNumber, columnNumber):
    self.type = type
    self.string = string
    self.lineNumber = lineNumber
    self.columnNumber = columnNumber

# 3. Exceptions
# -------------

class ParseError(Exception):
  def __init__(self, phase):
    self.phase = phase

  def GetPhase(self):
    return self.phase

# 4. Lexer
# --------

class Lexer:
  def __init__(self):
    self.lineNumber = 0
    self.columnNumber = 0
    self.line = None
    self.tokens = []

  def ImportFile(self, fileName):
    try:
      file = open(fileName, "r")
    except:
      sys.stderr.write("Unable to open file {0:s}\n".format(fileName))
      sys.stderr.flush()
      return 1
    try:
      self.ReadTokensInFile(file)
    except ParseError as exception:
      sys.stderr.write("{0:s} aborted\n".format(exception.GetPhase()))
      sys.stderr.flush()
      return 1
    return 0

  def ReadTokensInFile(self, file):
    self.lineNumber = 0
    self.tokens = []
    for self.line in file:
      self.lineNumber += 1
      self.ReadTokensInLine()

  def ReadTokensInLine(self):
    self.columnNumber = 1
    self.line = self.line.rstrip('\r\n')
    sourceLine = self.line[:]
    while self.line:
      if re.match(r"^[ \t]+", self.line):
        result = re.match(r"^[ \t]+", self.line)
        string = result.group()
        self.line = self.line[len(string):]
        self.columnNumber += len(string)
      elif self.line[0]=="(":
        self.ReadTokenInLine(Token.LPAR, "(")
      elif self.line[0]==")":
        self.ReadTokenInLine(Token.RPAR, ")")
      elif self.line[0]==":":
        self.ReadTokenInLine(Token.COLON, ":")
      elif self.line[0]==";":
        self.ReadTokenInLine(Token.SEMICOLON, ";")
      elif self.line[0]==",":
        self.ReadTokenInLine(Token.COMMA, ",")
      elif self.line[0]=="=":
        self.ReadTokenInLine(Token.EQUAL, "=")
      elif self.line[0]=="+":
        self.ReadTokenInLine(Token.PLUS, "+")
      elif self.line[0]=="-":
        self.ReadTokenInLine(Token.MINUS, "-")
      elif self.line[0]=="*":
        self.ReadTokenInLine(Token.STAR, "*")
      elif self.line[0]=="/":
        self.ReadTokenInLine(Token.SLASH, "/")
      elif re.match(r"^model", self.line):
        self.ReadTokenInLine(Token.KEYWORD_MODEL, "model")
      elif re.match(r"^end", self.line):
        self.ReadTokenInLine(Token.KEYWORD_END, "end")
      elif re.match(r"^state", self.line):
        self.ReadTokenInLine(Token.KEYWORD_STATE, "state")
      elif re.match(r"^initial", self.line):
        self.ReadTokenInLine(Token.KEYWORD_INITIAL, "initial")
      elif re.match(r"^transition", self.line):
        self.ReadTokenInLine(Token.KEYWORD_TRANSITION, "transition")
      elif re.match(r"^parameter", self.line):
        self.ReadTokenInLine(Token.KEYWORD_PARAMETER, "parameter")
      elif re.match(r"^reward", self.line):
        self.ReadTokenInLine(Token.KEYWORD_REWARD, "reward")
      elif re.match(r"^[a-zA-Z_][a-zA-Z0-9_]*", self.line):
        result = re.match(r"^[a-zA-Z_][a-zA-Z0-9_]*", self.line)
        string = result.group()
        self.ReadTokenInLine(Token.IDENTIFIER, string)
      elif re.match(r"^(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?", self.line):
        result = re.match(r"(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?", self.line)
        string = result.group()
        self.ReadTokenInLine(Token.NUMBER, string)
      else:
        message = "Line {0:d}, column {1:d}: unknown token\n".format(self.lineNumber, self.columnNumber)
        self.RaiseError(message)

  def ReadTokenInLine(self, type, string):
    token = Token(type, string, self.lineNumber, self.columnNumber)
    self.tokens.append(token)
    self.line = self.line[len(string):]
    self.columnNumber += len(string)
    return token

  def RaiseError(self, message):
    sys.stderr.write(message)
    sys.stderr.flush()
    raise ParseError("Lexical analysis")

  def DumpTokens(self, outputFile):
    for token in self.tokens:
      self.DumpToken(token, outputFile)

  def DumpToken(self, token, outputFile):
    outputFile.write("{0:d},{1:d}: {2:s} ({3:d})\n".format(token.lineNumber, token.columnNumber, token.string, token.type))

# 4. Parser
# ---------

class Parser:
  def __init__(self):
    self.lexer = Lexer()
    self.tokenIndex = 0
    self.model = None

  def ImportModel(self, fileName):
    try:
      file = open(fileName, "r")
    except:
      sys.stderr.write("Unable to open file {0:s}\n".format(fileName))
      sys.stderr.flush()
      return None
    try:
      self.lexer.ReadTokensInFile(file)
    except ParseError as exception:
      sys.stderr.write("{0:s} aborted\n".format(exception.GetPhase()))
      sys.stderr.flush()
      return None
    self.tokenIndex = 0
    try:
      self.ReadModel()
    except ParseError as exception:
      sys.stderr.write("{0:s} aborted\n".format(exception.GetPhase()))
      sys.stderr.flush()
      return None
    return self.model

  def ReadModel(self):
    token = self.GetGivenToken(Token.KEYWORD_MODEL)
    token = self.GetGivenToken(Token.IDENTIFIER)
    self.model = MarkovModel(token.string)
    while True:
      token = self.GetCurrentToken()
      if token.type==Token.KEYWORD_END:
        break
      self.ReadDeclaration()
    token = self.GetGivenToken(Token.KEYWORD_END)

  def ReadDeclaration(self):
    token = self.GetCurrentToken()
    if token.type==Token.KEYWORD_STATE:
      self.ReadStateDeclaration()
    elif token.type==Token.KEYWORD_TRANSITION:
      self.ReadTransitionDeclaration()
    elif token.type==Token.KEYWORD_PARAMETER:
      self.ReadParameterDeclaration()
    elif token.type==Token.KEYWORD_REWARD:
      self.ReadRewardDeclaration()
    else:
      message = "Line {0:d}, column {1:d}: unexpected token\n".format(token.lineNumber, token.columnNumber) 
      self.RaiseError(message)

  def ReadStateDeclaration(self):
    token = self.GetGivenToken(Token.KEYWORD_STATE)
    token = self.GetGivenToken(Token.IDENTIFIER)
    state = self.model.LookForState(token.string)
    if state!=None:
      message = "Line {0:d}, column {1:d}: already defined state {2:s}\n".format(token.lineNumber, token.columnNumber, token.string) 
      self.RaiseError(message)
    state = self.model.NewState(token.string)
    token = self.GetCurrentToken()
    if token.type==Token.LPAR:
      token = self.GetGivenToken(Token.LPAR)
      token = self.GetGivenToken(Token.KEYWORD_INITIAL)
      token = self.GetGivenToken(Token.EQUAL)
      token = self.GetGivenToken(Token.NUMBER)
      state.SetInitialProbability(float(token.string))
      token = self.GetGivenToken(Token.RPAR)
    token = self.GetGivenToken(Token.SEMICOLON)

  def ReadTransitionDeclaration(self):
    token = self.GetGivenToken(Token.KEYWORD_TRANSITION)
    token = self.GetGivenToken(Token.IDENTIFIER)
    sourceState = self.model.LookForState(token.string)
    if sourceState==None:
      message = "Line {0:d}, column {1:d}: undefined state {2:s}\n".format(token.lineNumber, token.columnNumber, token.string) 
      self.RaiseError(message)
    token = self.GetGivenToken(Token.IDENTIFIER)
    targetState = self.model.LookForState(token.string)
    if targetState==None:
      message = "Line {0:d}, column {1:d}: undefined state {2:s}\n".format(token.lineNumber, token.columnNumber, token.string) 
      self.RaiseError(message)
    if sourceState==targetState:
      message = "Line {0:d}, column {1:d}: source and target states must be different\n".format(token.lineNumber, token.columnNumber) 
      self.RaiseError(message)
    rateExpression = self.ReadExpression()
    self.model.NewTransition(sourceState, targetState, rateExpression)
    token = self.GetGivenToken(Token.SEMICOLON)

  def ReadParameterDeclaration(self):
    token = self.GetGivenToken(Token.KEYWORD_PARAMETER)
    token = self.GetGivenToken(Token.IDENTIFIER)
    parameter = self.model.LookForParameter(token.string)
    if parameter!=None:
      message = "Line {0:d}, column {1:d}: already defined parameter {2:s}\n".format(token.lineNumber, token.columnNumber, token.string) 
      self.RaiseError(message)
    parameter = self.model.NewParameter(token.string)
    token = self.GetGivenToken(Token.EQUAL)
    definition = self.ReadExpression()
    parameter.SetDefinition(definition)
    token = self.GetGivenToken(Token.SEMICOLON)

  def ReadRewardDeclaration(self):
    token = self.GetGivenToken(Token.KEYWORD_REWARD)
    token = self.GetGivenToken(Token.IDENTIFIER)
    reward = self.model.LookForReward(token.string)
    if reward!=None:
      message = "Line {0:d}, column {1:d}: already defined reward {2:s}\n".format(token.lineNumber, token.columnNumber, token.string) 
      self.RaiseError(message)
    reward = self.model.NewReward(token.string)
    token = self.GetGivenToken(Token.EQUAL)
    while True:
      token = self.GetGivenToken(Token.IDENTIFIER)
      state = self.model.LookForState(token.string)
      if state==None:
        message = "Line {0:d}, column {1:d}: undefined state {2:s}\n".format(token.lineNumber, token.columnNumber, token.string) 
        self.RaiseError(message)
      token = self.GetGivenToken(Token.COLON)
      value = self.ReadExpression()
      reward.SetValueOfState(state, value)
      token = self.GetCurrentToken()
      if token.type!=Token.COMMA:
        break
      self.tokenIndex += 1
    token = self.GetGivenToken(Token.SEMICOLON)

  def ReadExpression(self):
    token = self.GetCurrentToken()
    if token.type==Token.IDENTIFIER:
      self.tokenIndex += 1
      return Expression.NewParameterReference(token.string)
    if token.type==Token.NUMBER:
      self.tokenIndex += 1
      return Expression.NewReal(float(token.string))
    token = self.GetGivenToken(Token.LPAR)
    token = self.GetCurrentToken()
    if token.type==Token.PLUS:
      self.tokenIndex += 1
      expression = Expression.NewAdd()
    elif token.type==Token.MINUS:
      self.tokenIndex += 1
      expression = Expression.NewSub()
    elif token.type==Token.STAR:
      self.tokenIndex += 1
      expression = Expression.NewMul()
    elif token.type==Token.SLASH:
      self.tokenIndex += 1
      expression = Expression.NewDiv()
    else:
      message = "Line {0:d}, column {1:d}: unknown operator {2:s}\n".format(token.lineNumber, token.columnNumber, token.string) 
      self.RaiseError(message)
    while True:
      token = self.GetCurrentToken()
      if token.type==Token.RPAR:
        break
      childExpression = self.ReadExpression()
      expression.AddChild(childExpression)
    token = self.GetGivenToken(Token.RPAR)
    return expression

  def LookForCurrentToken(self):
    if self.tokenIndex<0 or self.tokenIndex>len(self.lexer.tokens):
      return None
    return self.lexer.tokens[self.tokenIndex]

  def GetCurrentToken(self):
    token = self.LookForCurrentToken()
    if token==None:
      self.RaiseError("Unexpected end of file\n")
    return token

  def GetGivenToken(self, tokenType):
    token = self.GetCurrentToken()
    if token.type!=tokenType:
      message = "Line {0:d}, column {1:d}: unexpected token ({2:s})\n".format(token.lineNumber, token.columnNumber, token.string) 
      self.RaiseError(message)
    self.tokenIndex += 1
    return token

  def RaiseError(self, message):
    sys.stderr.write(message)
    sys.stderr.flush()
    raise ParseError("Syntax analysis")

# 5. Checker
# ----------

class Checker:
  def __init__(self):
    self.model = None
    self.currentState = None
    self.currentTransition = None
    self.currentParameter = None
    self.currentReward = None
    self.error = 0

  def CheckModel(self, model):
    self.model = model
    self.error = 0
    self.CheckParameterReferences()
    self.CheckInitialProbabilities()
    return self.error

  def CheckParameterReferences(self):
    for transition in self.model.GetTransitions():
      self.currentTransition = transition
      self.CheckParameterReferencesInExpression(transition.GetRateExpression())
    self.currentTransition = None
    for parameter in self.model.GetParameters():
      self.currentParameter = parameter
      self.CheckParameterReferencesInExpression(parameter.GetDefinition())
    self.currentParameter = None
    for reward in self.model.GetRewards():
      self.currentReward = reward
      for state in reward.GetStates():
        self.CheckParameterReferencesInExpression(reward.GetValueOfState(state))
    self.currentReward = None

  def CheckParameterReferencesInExpression(self, expression):
    if expression.GetType()==Expression.PARAMETER:
      name = expression.GetParameterName()
      parameter = self.model.LookForParameter(name)
      if parameter==None:
        if self.currentTransition!=None:
          message = "In transition from {0:s} to {1:s}: ".format(self.currentTransition.GetSourceState().GetName(), self.currentTransition.GetTargetState().GetName())
        elif self.currentParameter!=None:
          message = "In declaration of parameter {0:s}: ".format(self.currentParameter.GetName())
        elif self.currentReward!=None:
          message = "In declaration of reward {0:s}: ".format(self.currentReward.GetName())
        message += "Undefined parameter reference {0:s}\n".format(name)
        self.RaiseError(message)
        return
      expression.SetParameter(parameter)
      return
    if expression.GetType()==Expression.REAL:
      return
    for childExpression in expression.GetChildren():
      self.CheckParameterReferencesInExpression(childExpression)

  def CheckInitialProbabilities(self):
    sumOfProbabilities = 0
    for state in self.model.GetStates():
      probability = state.GetInitialProbability()
      if probability<0.0 or probability>1.1:
        message = "Initial probability of state {0:s} out of range".format(state.GetName())
        self.RaiseError(message)
      sumOfProbabilities += probability
    if sumOfProbabilities!=1.0:
      message = "Initial probabilities do not sum to 1"
      self.RaiseError(message)

  def RaiseError(self, message):
    sys.stderr.write(message)
    sys.stderr.flush()
    self.error += 1

# 6. Printer
# ----------

class Printer:
  def __init__(self):
    self.indentation = "\t"

  def ExportModel(self, model, fileName):
    try:
      file = open(fileName, "w")
    except:
      sys.stderr.write("Unable to open file {0:s}\n".format(fileName))
      sys.stderr.flush()
      sys.exit()
    self.PrintModel(model, file)
    file.flush()
    file.close()

  def PrintModel(self, model, file):
    file.write("model {0:s}\n".format(model.GetName()))
    for state in model.GetStates():
      self.PrintState(state, file)
    for transition in model.GetTransitions():
      self.PrintTransition(transition, file)
    for reward in model.GetRewards():
      self.PrintReward(reward, file)
    for parameter in model.GetParameters():
      self.PrintParameter(parameter, file)
    file.write("end")

  def PrintState(self, state, file):
    file.write("{0:s}state {1:s}".format(self.indentation, state.GetName()))
    if state.GetInitialProbability()!=0.0:
      file.write("(initial = {0:g})".format(state.GetInitialProbability()))
    file.write(";\n")

  def PrintTransition(self, transition, file):
    file.write("{0:s}transition ".format(self.indentation))
    file.write("{0:s}".format(transition.GetSourceState().GetName()))
    file.write(" ")
    file.write("{0:s}".format(transition.GetTargetState().GetName()))
    file.write(" ")
    self.PrintExpression(transition.GetRateExpression(), file)
    file.write(";\n")

  def PrintReward(self, reward, file):
    file.write("{0:s}reward ".format(self.indentation))
    file.write("{0:s}".format(reward.GetName()))
    file.write(" =\n")
    first = True
    for state in reward.GetStates():
      if first:
        first = False
      else:
        file.write(",\n")
      file.write("{0:s}{0:s} ".format(self.indentation))
      file.write("{0:s}: ".format(state.GetName()))
      self.PrintExpression(reward.GetValueOfState(state), file)
    file.write(";\n")

  def PrintParameter(self, parameter, file):
    file.write("{0:s}parameter ".format(self.indentation))
    file.write("{0:s}".format(parameter.GetName()))
    file.write(" = ")
    self.PrintExpression(parameter.GetDefinition(), file)
    file.write(";\n")

  def PrintExpression(self, expression, file):
    if expression.type==Expression.PARAMETER:
      file.write("{0:s}".format(expression.GetParameterName()))
    elif expression.type==Expression.REAL:
      file.write("{0:g}".format(expression.GetValue()))
    elif expression.type==Expression.ADD:
      self.PrintParenthesedExpression(expression, "+", file)
    elif expression.type==Expression.SUB:
      self.PrintParenthesedExpression(expression, "-", file)
    elif expression.type==Expression.MUL:
      self.PrintParenthesedExpression(expression, "*", file)
    elif expression.type==Expression.DIV:
      self.PrintParenthesedExpression(expression, "/", file)

  def PrintParenthesedExpression(self, expression, operator, file):
    file.write("({0:s}".format(operator))
    for childExpression in expression.GetChildren():
      file.write(" ")
      self.PrintExpression(childExpression, file)
    file.write(")")

